;; Initialization and network definition for load flow calculation
;; in Cartesian coordinates.   **** This version for hardware ****
;; Each processor plays three roles:
;;    1) Calculates c, d and sends to partner at other end of wire
;;    2) Recieves c, d from partner and participates in sums for C, D
;;    3) Represents its node, calculates new u, v (redundantly)

;; This version reads Gii and Bii as data rather than calculating

(*proclaim '(type fixnum mxn mxe))

(defvar mxn 0)   ;; Maximum number of nodes = grid size in 0 coord
(defvar mxe 0)   ;; Maximum degree of nodes = grid size in 1 coord

;; Network representation - other pvar's needed for calculation are
;; introduced by the files containing the solution routines.

(*proclaim
 '(type single-float-pvar p q u u2 v w gm bm cm dm C D f a wk))

(*defvar p    (!! 0.0))    ; p for node
(*defvar q    (!! 0.0))    ; q for node
(*defvar u    (!! 0.0))    ; u for node (generators) u^2 = v^2 + w^2
(*defvar u2   (!! 0.0))    ; u^2 for node
(*defvar v    (!! 0.0))    ; v for node
(*defvar w    (!! 0.0))    ; w for node
(*defvar gm   (!! 0.0))    ; g for line
(*defvar bm   (!! 0.0))    ; b for line
(*defvar cm   (!! 0.0))    ; cm = gm*v - bm*w
(*defvar dm   (!! 0.0))    ; dm = gm*w + bm*v
(*defvar C    (!! 0.0))    ; C = SUM cm over adjacent nodes
(*defvar D    (!! 0.0))    ; D = SUM dm over adjacent nodes
(*defvar f    (!! 0.0))    ; 1/Ykk = f + j*a
(*defvar a    (!! 0.0))
(*defvar wk   (!! 0.0))    ; permanent working variable

(*proclaim '(type (unsigned-pvar *current-send-address-length*) to))

(*defvar to (!! 0)) ;; Pointer to partner

(*proclaim '(type (unsigned-pvar 20) myx myy))

(*defvar myx (self-address-grid!! (!! 0))) ;; My node number
(*defvar myy (self-address-grid!! (!! 1))) ;; My position in the group


(*proclaim '(type Boolean-pvar act snd gen clc))

(*defvar act nil!!)        ;; Active in graph rep
(*defvar snd nil!!)        ;; I am next proc to use as sender for my node
(*defvar gen nil!!)        ;; I am a generator
(*defvar clc nil!!)        ;; I calculate new values (not a slack bus)

;; Set up CM with virtual dimensions from mxn mxe

(defun setup ()
 (*cold-boot :initial-dimensions (list mxn mxe))
)

(defvar NODES 0)   ; Actual number of nodes
(defvar EDGES 0)   ; Actual number of edges (transmission lines)

;; Enter edge from na to nt and symmetric edge from nt to na
;; with coefficients gx, bx.  Reports edges which cannot be
;; entered.

(defun enter-edge (na nt gx bx)
 (let ((rc 0))
      (declare (type fixnum rc))
   (if (and (*or (and!! (=!! myx (!! nt)) snd))
	    (*or (and!! (=!! myx (!! na)) snd))) ;; Receiver and sender OK
    (progn
     (incf EDGES 2)
     (*when (and!! (=!! myx (!! nt)) snd)
       (do-for-selected-processors (c) ; exactly one processor
         (*setf (pref act c) t)  (setq rc c)
         (*setf (pref gm c) gx)   (*setf (pref bm c) bx)
         (if (< (pref myy c) (1- mxe)) (*news snd snd 0 1))
         (*setf (pref snd c) nil))
      )  ; end *when
     (*when (and!! (=!! myx (!! na)) snd)
       (do-for-selected-processors (c) ; exactly one processor
         (*setf (pref act c) t)
	 (*setf (pref to c) rc)  (*setf (pref to rc) c)
         (*setf (pref gm c) gx)   (*setf (pref bm c) bx)
	 (if (< (pref myy c) (1- mxe)) (*news snd snd 0 1))
	 (*setf (pref snd c) nil)))
     ) ; end progn
    (format t "~%No room for edge ~D to ~D.~%" na nt)
   ))  ; end if (and ...), end let
)  ; end enter-edge

;; Enter Gnn and Bnn reciprocals as f and g in all relevant processors

(defun enter-diag (n gx bx)
  (*when (=!! myx (!! n))
    (*set f (!! gx)) (*set a (!! bx))
    (*set wk (+!! (*!! f f) (*!! a a)))
    (*set f (/!! f wk))  (*set a (-!! (/!! a wk))))
) ; end enter-diag

;; Read graph description from stream and install - Cm should be setup
;; suitably.  Ignores edges which won't fit.

(defun loadg (strm)
 (*set gm (!! 0.0)) (*set bm (!! 0.0))
 (*set  act nil!!)  (*set snd nil!!)
 (*when (=!! myy (!! 0)) (*set snd t!!))
 (setq EDGES 0)
 (do ((n (read strm) (read strm)))
     ((< n 0))
   (do ((nt (read strm) (read strm))
        (gx (read strm) (read strm)) (bx (read strm) (read strm)))
       ((< nt 0))
     (if (/= n nt) (enter-edge n nt gx bx) (enter-diag n gx bx))
    )  ; end do nt, gx, bx
  )  ; end do n
 (*when act (setq NODES (1+ (*max myx))))
 (list NODES EDGES)
)  ; end loadg

;; Load node data for generators

(defun loadgen (strm)
 (do ((n (read strm) (read strm)))
     ((< n 0))
  (let ((px (read strm)) (ux (read strm)))
    (*when (=!! myx (!! n))
       (*setf p (!! px)) (*setf u (!! ux)) (*set gen t!!)))
 ) ; end do
)

;; Load node data for loads

(defun loadld (strm)
 (do ((n (read strm) (read strm)))
     ((< n 0))
  (let ((px (read strm)) (qx (read strm)))
    (*when (=!! myx (!! n))
       (*setf p (!! px)) (*setf q (!! qx)) (*set gen nil!!)))
 ) ; end do
)

;; Initialize for v, w, derived quantities for obtaining solution
;; Loads are initialized with v = lv;

(defun initvw (lv)
  (*set w (!! 0.0))  (*set u2 (*!! u u))
  (*if gen (*set v u) (*set v (!! lv)))
  (*set C (!! 0))  (*set D (!! 0))
  (*set clc (and!! act (plusp!! myx)))
 ; (*when act ;; Omits calculation of diagonal elements
 ;   (*pset :no-collisions gm f to)  ;; For calculation of 1/Ykk
 ;   (*pset :no-collisions bm a to)
 ;   (*set f (reduce-and-spread!! f '+!! 1))
 ;   (*set a (reduce-and-spread!! a '+!! 1))
 ;   (*set wk (+!! (*!! f f) (*!! a a)))
 ;   (*set f (-!! (/!! f wk)))  (*set a (/!! a wk)))
)


;; Replicate portion of current graph consisting of first n nodes.
;; Makes no connections between old and new components.
;; Intended for situations in which nodes are 0, ..., NODES-1 (contiguous).
;; The portion replicated should be isolated from the rest of the graph.

(defun repl (n)
 (let ((sndn 0) (dltaE 0))
      (declare (type fixnum sndn dltaE))
   (*when snd (setq sndn (*min myx)))
   (when (or (null snd) (> (+ n NODES) mxn))
         (format t "Can't replicate the graph.~%")
	 (return-from repl))
   (*let ((buddy (!! 0)))
     (*when (<!! myx (!! n))
        (*when act (setq dltaE (*sum (!! 1))))
	(*set buddy (cube-from-grid-address!! (+!! myx (!! NODES)) myy)))
     (*when (<!! myx (!! n))
       (*pset :no-collisions act act buddy)
       (*pset :no-collisions snd snd buddy)
       (*pset :no-collisions gen gen buddy)
       (*pset :no-collisions (pref!! buddy to) to buddy)
       (*pset :no-collisions gm gm buddy)
       (*pset :no-collisions bm bm buddy)
       (*pset :no-collisions p  p  buddy)
       (*pset :no-collisions q  q  buddy)
       (*pset :no-collisions u  u  buddy)
     )
     (setq EDGES (+ dltaE EDGES))  (setq NODES (+ n NODES))
   )  ; end *let
 (list NODES EDGES)
 )  ; end let
)  ; end double

(defvar file "gtest.dat")      ; The current data file (a string)

;; Load graph from data file whose name is in file.

(defun sg ()
  (with-open-file (strmx file) (loadg strmx) (loadgen strmx) (loadld strmx))
)

;; Load graph and answers from file whose name is in file

(defun sga ()
  (with-open-file (strmx file)
    (loadg strmx) (loadgen strmx) (loadld strmx) (ldans strmx))
)

;; Load answers from strm, all u's then all v's, in order of node number

(defvar vans nil)
(defvar wans nil)

(defun ldans (strm) (setq vans (ldvec strm)) (setq wans (ldvec strm)))

;; Read vector with NODES entries

(defun ldvec (strm)
  (let ((av (make-array NODES :element-type 'single-float
			      :initial-element 0.0)))
    (dotimes (i NODES) (setf (aref av i) (read strm)))
    av)
)

;; Initialize for v, w, derived quantities for obtaining solution
;; using answers as loaded by ldans

(defun initvwans ()
  (*set w (!! 0.0))  (*set u2 (*!! u u))
  (dotimes (n NODES)
    (*when (=!! myx (!! n))
      (*set v (!! (aref vans n)))  (*set w (!! (aref wans n)))))
  (*set C (!! 0))  (*set D (!! 0))
  (*set clc (and!! act (plusp!! myx)))
)
